use lasso::{Rodeo, Spur};
use logos::{Lexer, Logos, Skip};

#[derive(Default)]
pub struct Extras {
    int: Rodeo,
}

#[derive(Logos, Clone, Copy, Debug, PartialEq, Eq, Hash)]
#[logos(skip r"[ \t\n\f]+")]
#[logos(skip r"owo .*")]
#[logos(extras = Extras)]
#[rustfmt::skip]
pub enum Token {
    #[token("(")] LParen,
    #[token(")")] RParen,
    #[token(",")] Comma,
    #[token(";")] Semicolon,
    #[token("`")] Tick,
    #[token("+")] Plus,
    #[token("-")] Minus,
    #[token("*")] Star,
    #[token("/")] Slash,
    #[token("\\")] Backslash,
    #[token("^")] Caret,
    #[token("<")] Lt,
    #[token(">")] Gt,
    #[token("=")] Eq,
    #[token("ain't")] Aint,
    #[token("=:")] Assign,

    #[token("top")] Top,
    #[token("bottom")] Bottom,
    #[token("functio")] Functio,
    #[token("unless")] Unless,
    #[token("result")] Result,

    #[regex(r"\p{XID_Start}(\p{XID_Continue}|-)*", intern)]
    Ident(Spur),

    #[regex(r"-?[0-9]+", integer)]
    Int((bool, u64)),

    #[doc(hidden)]
    #[token("uwu top", mlcomment)]
    Invalid,
}

#[inline]
fn intern(lexer: &mut Lexer<Token>) -> Spur {
    lexer.extras.int.get_or_intern(lexer.slice())
}

#[inline]
fn integer(lexer: &Lexer<Token>) -> Option<(bool, u64)> {
    let (sign, num) = match lexer.slice().strip_prefix('-') {
        Some(n) => (true, n),
        None => (false, lexer.slice()),
    };

    Some((sign, num.parse().ok()?))
}

#[inline]
fn mlcomment(lexer: &mut Lexer<Token>) -> Skip {
    let mut count = 1;
    loop {
        match lexer.next() {
            Some(Ok(Token::Bottom)) if count == 1 => break,
            Some(Ok(Token::Top)) => count += 1,
            Some(Ok(Token::Bottom)) => count -= 1,
            Some(_) => (),
            None => break,
        }
    }

    Skip
}
